/*
 * Server code generation.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "common.h"
#include "config.h"
#include "declaration.h"
#include "message.h"
#include "server.h"
#include "templates.h"



/* Generate each component operation signature or "function prototype". */

void write_server_interface_signature(struct signature *sig,
                                      enum signature_role role, FILE *fp,
                                      struct interface *iface)
{
  write_interface_signature(sig, role, SERVER_ROLE, fp, iface);
}

/* Generate signature declarations for each operation. */

void write_server_signature(struct signature *sig, enum signature_role role,
                            FILE *fp, struct interface *iface)
{
  if (have_attribute(sig->attributes, "completion"))
    write_server_initiation_signature(sig, role, fp, iface);
  else
    write_server_wrapper_signature(sig, role, fp, iface);

  /* Write the appropriate completion signature even for use by conventional
     operations. */

  write_server_completion_signature(sig, role, fp, iface);
}

/* Generic signature generation. */

static void _write_server_signature(struct signature *sig,
                                    enum signature_role role,
                                    FILE *fp, struct interface *iface,
                                    const char *prefix)
{
  char *opname = get_operation_name(iface, sig);

  fprintf(fp, server_function_signature, prefix, opname, iface->name);
  fputs(get_signature_terminator(role), fp);

  free(opname);
}

/* Generate operation wrapper function details used by the server. */

void write_server_wrapper_signature(struct signature *sig, enum signature_role role,
                            FILE *fp, struct interface *iface)
{
  _write_server_signature(sig, role, fp, iface, "wrap");
}

void write_server_initiation_signature(struct signature *sig,
                                       enum signature_role role,
                                       FILE *fp, struct interface *iface)
{
  _write_server_signature(sig, role, fp, iface, "initiate");
}

void write_server_completion_signature(struct signature *sig,
                                       enum signature_role role,
                                       FILE *fp, struct interface *iface)
{
  char *opname = get_operation_name(iface, sig);

  /* Generate a signature featuring an initiator reference (for genuine
     decoupled completions) and only "out" and "inout" parameters. */

  int completion = have_attribute(sig->attributes, "completion");

  char *prologue = completion ?
                   server_completion_function_signature_prologue :
                   server_completion_optional_function_signature_prologue;

  fprintf(fp, prologue, opname, iface->name);

  /* Continue from the initial endpoint parameter if a genuine completion. */

  write_parameters(sig->parameters, fp, SIGNATURE_ROLE, COMPLETION_ROLE,
                   OUT_PARAMETER, completion);

  fputs(")", fp);
  fputs(get_signature_terminator(role), fp);

  free(opname);
}

/* Generate function source code for each operation, with the generated function
   unpacking a message, calling the actual operation and repacking the
   results. */

void write_server_function(struct signature *sig, FILE *fp, struct interface *iface)
{
  if (have_attribute(sig->attributes, "completion"))
  {
    write_server_initiation_signature(sig, DEFINITION_ROLE, fp, iface);
    write_server_initiation_function_body(sig->parameters, fp, iface, sig);
    fputs(END_FUNCTION, fp);
  }
  else
  {
    write_server_wrapper_signature(sig, DEFINITION_ROLE, fp, iface);
    write_server_wrapper_function_body(sig->parameters, fp, iface, sig);
    fputs(END_FUNCTION, fp);
  }

  /* Write the appropriate completion function even for use by conventional
     operations. */

  write_server_completion_signature(sig, DEFINITION_ROLE, fp, iface);
  write_server_completion_function_body(sig->parameters, fp, iface, sig);
  fputs(END_FUNCTION, fp);
}

/* Generate a function body corresponding to an operation for server use. */

void write_server_wrapper_function_body(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  char *opname = get_operation_name(iface, sig);

  /* Generate the prologue. */

  fputs(server_function_body_prologue, fp);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Generate variable declarations. */

  write_declarations(param, ANY_PARAMETER, ANY_CLASS, fp);

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(input_words, output_words, opname, fp);

  /* Handle superfluous message structures. */

  write_superfluous_message_cast(fp, input_words, input_items, output_words, output_items);

  /* Unpack each word and item from the message into variables. */

  write_input_initialisation(param, fp, opname, input_words, input_items);

  /* Invoke the actual operation using the variables. */

  write_server_function_call(param, fp, iface, sig, ANY_PARAMETER);

  /* Pack the outputs into the message and perform any other housekeeping. */

  write_output_initialisation(param, fp, opname, output_words, output_items, GENERAL_FUNCTION_ROLE);

  /* Return the success of the reply operation. */

  fputs(server_function_body_epilogue, fp);

  /* Free allocated strings. */

  free(opname);
}

/* Generate a function body corresponding to the initiation of an operation. */

void write_server_initiation_function_body(struct parameter *param, FILE *fp,
                                           struct interface *iface,
                                           struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  char *opname = get_operation_name(iface, sig);

  /* Generate the prologue. */

  fputs(server_function_body_prologue, fp);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Generate variable declarations for "in" and "inout" parameters. */

  write_declarations(param, IN_PARAMETER, ANY_CLASS, fp);

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(input_words, 0, opname, fp);

  /* Handle superfluous message structures. */

  write_superfluous_message_cast(fp, input_words, input_items, 0, 0);

  /* Unpack each word and item from the message into variables. */

  write_input_initialisation(param, fp, opname, input_words, input_items);

  /* Invoke the actual operation using the variables. */

  write_server_function_call(param, fp, iface, sig, IN_PARAMETER);

  /* Return the success of the operation restoring buffer registers. */

  fputs(server_initiation_function_body_epilogue, fp);

  /* Free allocated strings. */

  free(opname);
}

/* Generate a function body corresponding to the completion of an operation. */

void write_server_completion_function_body(struct parameter *param, FILE *fp,
                                           struct interface *iface,
                                           struct signature *sig)
{
  int input_words, input_items, output_words, output_items;
  char *opname = get_operation_name(iface, sig);

  /* Generate the prologue. */

  fputs(server_completion_function_body_prologue, fp);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(0, output_words, opname, fp);

  /* Initialise the message. */

  fprintf(fp, "\n  ipc_message_new(&msg);\n");

  /* Pack the outputs into the message and perform any other housekeeping. */

  write_output_initialisation(param, fp, opname, output_words, output_items, COMPLETION_ROLE);

  /* Send a response as a new message to the given endpoint.
     NOTE: The label 0 is employed but an error condition could be communicated. */

  if (have_attribute(sig->attributes, "completion"))
    fprintf(fp, server_completion_function_body_epilogue, "0", "_endp");

  /* Send a plain reply for optional/coupled completions. */

  else
    fputs(server_completion_optional_function_body_epilogue, fp);

  /* Free allocated strings. */

  free(opname);
}

/* Generate a cast for superfluous message structures. */

void write_superfluous_message_cast(FILE *fp, int input_words, int input_items,
                                    int output_words, int output_items)
{
  if (!input_words && !input_items && !output_words && !output_items)
    fputs(server_function_body_unused_message, fp);
}

/* Generate the initialisation of input words and items. */

void write_input_initialisation(struct parameter *param, FILE *fp,
                                const char *opname, int input_words,
                                int input_items)
{
  if (input_words)
  {
    write_accessor_initialisation(IN_PARAMETER, SERVER_ROLE, GENERAL_FUNCTION_ROLE,
                                  opname, fp);
    write_message_access(param, SERVER_ROLE, GENERAL_FUNCTION_ROLE, IN_PARAMETER,
                         WORD_CLASS, fp);
  }

  if (input_items)
  {
    fputs("\n", fp);
    write_message_access(param, SERVER_ROLE, GENERAL_FUNCTION_ROLE, IN_PARAMETER,
                         ITEM_CLASS, fp);
  }
}

/* Generate the initialisation of output words and items. */

void write_output_initialisation(struct parameter *param, FILE *fp,
                                 const char *opname, int output_words,
                                 int output_items, enum function_role function)
{
  if (output_words)
  {
    write_accessor_initialisation(OUT_PARAMETER, SERVER_ROLE, function,
                                  opname, fp);
    write_message_access(param, SERVER_ROLE, function, OUT_PARAMETER,
                         WORD_CLASS, fp);
  }

  if (output_items)
  {
    fputs("\n", fp);
    write_message_access(param, SERVER_ROLE, function, OUT_PARAMETER,
                         ITEM_CLASS, fp);
  }
}

/* Generate an invocation of the actual server operation. */

void write_server_function_call(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig,
                                enum specifier specifier)
{
  int continuing = 0;
  char *name, *addr,
       *opname = get_signature_operation_name(iface, sig, SERVER_ROLE,
                                              ACCESS_ROLE, conf.language);

  fputs("\n  err = ", fp);

  /* Access a method when emitting C++. */

  if (conf.language == CPP_LANGUAGE)
    fprintf(fp, "_self->%s(", opname);

  /* Employ a function pointer via the object type otherwise. */

  else
  {
    fprintf(fp, "_self->iface->%s(_self->ref", opname);
    continuing = 1;
  }

  /* Generate the parameter list, employing addresses for output parameters. */

  write_parameters(param, fp, INVOCATION_ROLE, GENERAL_FUNCTION_ROLE, specifier,
                   continuing);
  fputs(");\n", fp);

  /* Emit post-invocation details. */

  fputs(server_function_body_call, fp);

  /* Free allocated strings. */

  free(opname);
}
